#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <ashd/req.h>
#include <ashd/resp.h>
#include <ashd/log.h>
#include <ashd/mt.h>
#include <ashd/mtio.h>

#include "statserve.h"

static struct source *sources = NULL;

static struct source *parsesource(char *arg)
{
    if(arg[strlen(arg) - 1] == '/') {
	return(mkfssrc(arg));
    } else {
	return(mkdbsrc(arg, NULL));
    }
}

static int cached(struct hthead *req, struct fileinfo f)
{
    char *hdr;
    time_t cdate;
    
    if((hdr = getheader(req, "If-Modified-Since")) != NULL) {
	cdate = parsehttpdate(hdr);
	return((cdate > 0) && !(cdate < f.mtime));
    }
    return(0);
}

static void serve(struct muth *muth, va_list args)
{
    vavar(struct hthead *, req);
    vavar(int, fd);
    FILE *out;
    struct source *src;
    struct fileinfo f;
    
    out = NULL;
    for(src = sources; src != NULL; src = src->next) {
	f = src->serve(src, req->rest);
	if(f.data != NULL)
	    break;
    }
    if(src == NULL) {
	simpleerror(fd, 404, "Resource not found", "The resource %s was not found", htmlquote(req->rest));
	goto out;
    }
    out = mtstdopen(fd, 1, 60, "r+");
    if(cached(req, f)) {
	fprintf(out, "HTTP/1.1 304 Not Modified\n");
	fprintf(out, "Content-Length: 0\n");
	fprintf(out, "\n");
    } else {
	fprintf(out, "HTTP/1.1 200 OK\n");
	fprintf(out, "Content-Type: %s\n", f.ctype);
	fprintf(out, "Content-Length: %zi\n", f.sz);
	fprintf(out, "Last-Modified: %s\n", fmthttpdate(f.mtime));
	fprintf(out, "\n");
	fwrite(f.data, 1, f.sz, out);
    }
    free(f.data);
    
out:
    if(out != NULL)
	fclose(out);
    else
	close(fd);
    freehthead(req);
}

static void usage(FILE *out)
{
    fprintf(out, "usage: statserve [-h] [-P PAGESIZE] SOURCE...\n");
}

static void listenloop(struct muth *muth, va_list args)
{
    vavar(int, lfd);
    int fd;
    struct hthead *req;
    struct source *src;
    
    while(1) {
	block(0, EV_READ, 0);
	if((fd = recvreq(lfd, &req)) < 0) {
	    if(errno != 0)
		flog(LOG_ERR, "recvreq: %s", strerror(errno));
	    break;
	}
	mustart(serve, req, fd);
	for(src = sources; src != NULL; src = src->next) {
	    if(src->idle)
		src->idle(src);
	}
    }
}

static void sigterm(int sig)
{
    shutdown(0, SHUT_RDWR);
}

static void closeall(void)
{
    struct source *src;
    
    for(src = sources; src != NULL; src = src->next) {
	if(src->close)
	    src->close(src);
    }
}

int main(int argc, char **argv)
{
    int c;
    struct source *last, *src;
    
    while((c = getopt(argc, argv, "+hP:")) >= 0) {
	switch(c) {
	case 'P':
	    dbpagesize = atoi(optarg);
	    break;
	case 'h':
	    usage(stdout);
	    return(0);
	default:
	    usage(stderr);
	    return(1);
	}
    }
    last = NULL;
    while(optind < argc) {
	if((src = parsesource(argv[optind++])) == NULL) {
	    closeall();
	    return(1);
	}
	if(!sources)
	    sources = src;
	if(last)
	    last->next = src;
	last = src;
    }
    if(!sources) {
	usage(stderr);
	closeall();
	return(1);
    }
    mustart(listenloop, 0);
    signal(SIGINT, sigterm);
    signal(SIGTERM, sigterm);
    ioloop();
    closeall();
    return(0);
}
